/*
 * Copyright (C) 2005 Universitat d'Alacant / Universidad de Alicante
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
#include <lttoolbox/Recognizer.H>

int const Recognizer::ANY_CHAR = -1;
int const Recognizer::ANY_TAG  = -2;

Recognizer::Recognizer()
{
}

Recognizer::~Recognizer()
{
  destroy();
}

Recognizer::Recognizer(Recognizer const &r)
{
  copy(r);
}


Recognizer &
Recognizer::operator =(Recognizer const &o)
{
  if(this != &o)
  {
    destroy();
    copy(o);
  }
  return *this;
}

void
Recognizer::copy(Recognizer const &o)
{
  fsm = o.fsm;
  idx = o.idx;
  idx_inv = o.idx_inv;
}

void
Recognizer::destroy()
{
}

int
Recognizer::encode(string const &symbol)
{
  map<string, int>::iterator it = idx.find(symbol);

  if(it != idx.end())
  {
    return it->second;
  }  
  else 
  {
    int newid = idx.size() + 1;
    idx[symbol] = newid;
    idx_inv[newid] = symbol;
    return newid;
  } 
}

string
Recognizer::decode(int const code)
{
  map<int, string>::iterator it = idx_inv.find(code);
  
  if(it != idx_inv.end())
  {
    return it->second;
  }
  else
  {
    return "";
  }
}

int
Recognizer::insertLemma(string const &lemma, int const where)
{
  int mystate = where;
  
  if(lemma == "")
  {
    mystate = fsm.insertSingleTransduction(ANY_CHAR, mystate);
    fsm.linkStates(mystate, mystate, ANY_CHAR);
  }
  else
  {
    for(unsigned int i = 0, limit = lemma.size(); i != limit; i++)
    {
      mystate = fsm.insertSingleTransduction(encode(lemma.substr(i, 1)), 
                                             mystate);
    }
  }  
  return mystate;
}

int 
Recognizer::insertTags(string const &tags, int const where)
{
  int mystate = where;

  for(unsigned int i = 0, j = 0, limit = tags.size(); j != limit ; j++)  
  {
    string symbol = "";
    
    if(tags[j] == '.')
    {
      symbol = tags.substr(i, j - i);
      i = j + 1;
    }
    else if(j == limit - 1)
    {
      symbol = tags.substr(i);
    }
    else
    {
      continue;
    }
    
    if(symbol == "*")
    {
      fsm.linkStates(mystate, mystate, ANY_TAG);
    }
    else
    {
      string const symbol2 = "<" + symbol + ">";
      mystate = fsm.insertSingleTransduction(encode(symbol2), mystate);
    }
  }
  
  return mystate;
}

int
Recognizer::insertWord(string const &lemma, string const &tags, int const where)
{
  int mystate = fsm.insertSingleTransduction(encode("^"), where);
  mystate =  insertLemma(lemma, mystate);
  mystate = insertTags(tags, mystate);
  return fsm.insertSingleTransduction(encode("$"), mystate);
}

int
Recognizer::insertBlank(int const where)
{
  return fsm.insertSingleTransduction(encode(" "), where);
}

int
Recognizer::getInitial()
{
  return fsm.getInitial();
}

int 
Recognizer::insertLastWord(string const &lemma, string const &tags, 
                           int const where, int const id)
{
  int mystate = insertWord(lemma, tags, where);
  fsm.setFinal(mystate);
  final_type[mystate] = id;
  return mystate;
}
