/*
 * Copyright (C) 2005 Universitat d'Alacant / Universidad de Alicante
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
#ifndef _RECOGNIZER_
#define _RECOGNIZER_

#include <lttoolbox/Transducer.H>

#include <list>

using namespace std;

/**
 * Class to represent a letter transducer during the dictionary compilation
 */
class Recognizer
{
private:
  Transducer fsm;
  map<string, int> idx;
  map<int, string> idx_inv;
  map<int, int> final_type;

  void copy(Recognizer const &o);
  void destroy();
  int insertLemma(string const &lemma, int const where);
  int insertTags(string const &lemma, int const where);  
  
  int encode(string const &symbol);
  string decode(int const code);


public:

  static int const ANY_CHAR;
  static int const ANY_TAG;
  /**
   * Constructor
   */
  Recognizer();

  /**
   * Destructor
   */
  ~Recognizer();

  /**
   * Copy constructor
   * @param r recognizer to be copied
   */
  Recognizer(Recognizer const &r);
  
  
  Recognizer & operator =(Recognizer const &o);
  
  /**
   * 
   */
  int insertWord(string const &lemma, string const &tags, int const where);

  int insertLastWord(string const &lemma, string const &tags, int const where,
                     int const id);
  int insertBlank(int const where);  
  int getInitial();
};

#endif
